#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "cmd.h"
#include "type.h"
#include "op.h"

obj_t * opstack;
obj_t * dictstack;
obj_t * gstatestack;

obj_t * systemdict;
obj_t * globaldict;
obj_t * userdict;

int compile = 0;
int trace = 0;
int tracecnt = 0;

void init()
{
	opstack = mkarr();
	dictstack = mkarr();
	gstatestack = mkarr();
	systemdict = mkdict();
	globaldict = mkdict();
	userdict = mkdict();

	arr_add(dictstack, systemdict);
	arr_add(dictstack, globaldict);
	arr_add(dictstack, userdict);

	op_init(systemdict);

	dict_set(systemdict, mkname("languagelevel"), mkint(2));
	dict_set(systemdict, mkname("globaldict"), globaldict);
}

void execobj(obj_t *o)
{
	assert(!compile);

	if(trace) {
		printf("%04d OP ", tracecnt);
		arr_add(opstack, o);
		Oprint();
		printf(" STACK ");
		arr_add(opstack, opstack);
		Oprint();
		printf("\n");

		tracecnt ++;
	}

	if(o->type == T_ARRAY && o->exec) {
		int j;
		for(j=0; j<o->arr.size; j++) {
			addobj(o->arr.v[j]);
		}
		return;
	}

	if(o->type != T_NAME || !o->exec) {
		arr_add(opstack, o);
		return;
	}

	if(!strcmp(o->s, "{")) {
		compile ++;
		obj_t *a = mkarr();
		a->exec = 1;
		arr_add(opstack, a);
		return;
	}

	int i;
	i = dictstack->arr.size - 1;
	while(i >= 0) {
		obj_t *d = dictstack->arr.v[i];
		obj_t *v = dict_lookup(d, o);
		if(v) {
			switch(v->type) {
			case T_OP:
				v->op();
				break;
			case T_ARRAY:
				if(v->exec) {
					int j;
					for(j=0; j<v->arr.size; j++) {
						addobj(v->arr.v[j]);
					}
					return;
				}
				/* fallthrough */
			default:
				arr_add(opstack, v);
			}
			return;
		}
		i --;
	}

	fprintf(stderr, "Not found: %s\n", o->s);
	exit(1);
}

void addobj(obj_t *o)
{
	if(compile) {
		if(o->type == T_NAME && o->exec) {
			if(!strcmp(o->s, "}")) {
				compile --;
			} else if(!strcmp(o->s, "{")) {
				compile ++;
			}
		}

		if(compile) arr_add(arr_peek(opstack), o);
	} else {
		execobj(o);
	}
}

void p_number(double v)
{
	addobj(mkreal(v));
}

void p_int(int v)
{
	addobj(mkint(v));
}

void p_token(const char *t)
{
	addobj(mkename(t));
}

void p_quote(const char *t)
{
	addobj(mkname(t));
}

void p_string(const char *t)
{
	addobj(mkstring(t));
}

void yerror(const char *msg)
{
	fprintf(stderr, "Parse error: %s\n", msg);
	exit(1);
}

