#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "cmd.h"
#include "op.h"
#include "gc.h"

extern void yylex();

int main(int argc, char *argv[])
{
	GC_INIT();

	int opt;
	char *p;
	double r1, r2;

	while((opt = getopt(argc, argv, "tx:y:b:")) != -1) {
		switch(opt) {
		case 'x':
			p = strchr(optarg, ',');
			if(!p) {
				fprintf(stderr, "Missing comma in arg\n");
				exit(1);
			}
			*p = 0;
			r1 = atof(optarg);
			r2 = atof(p + 1);
			devxscale = 2.0 / (r2 - r1);
			devxshift = -r1;
			break;
		case 'y':
			p = strchr(optarg, ',');
			if(!p) {
				fprintf(stderr, "Missing comma in arg\n");
				exit(1);
			}
			*p = 0;
			r1 = atof(optarg);
			r2 = atof(p + 1);
			devyscale = 2.0 / (r2 - r1);
			devyshift = -r1;
			break;
		case 't':
			trace = 1;
			break;
		case 'b':
			beziersegs = atoi(optarg);
			if(beziersegs < 2) {
				fprintf(stderr, "Invalid number of bezier segs\n");
				exit(1);
			}
			break;
		default:
			fprintf(stderr,
					"Usage:\n"
					" -x minx,maxx  - set X bounds for scaling\n"
					" -y miny,maxy  - set Y bounds for scaling\n"
					" -t            - enable trace\n"
					" -b <N>        - set bezier segs (by default 8)\n"
				   );
			exit(1);
		}
	}

	if(optind < argc) {
		fprintf(stderr, "superfluous arguments\n");
		exit(1);
	}

	init();
	yylex();

	fprintf(stderr, "X:%g,%g Y:%g,%g\n", minx, maxx, miny, maxy);
	return 0;
}
