%{
	#include "cmd.h"

	#define MAX_STR_CONST 1024
	char string_buf[MAX_STR_CONST];
	char *string_buf_ptr;

	int yywrap(void) {
		return 1;
	}
%}

%x str

%%

[0-9]*\.[0-9]+E[+-][0-9]+   p_number(atof(yytext));
-[0-9]*\.[0-9]+E[+-][0-9]+  p_number(atof(yytext));
[0-9]*\.[0-9]+E[0-9]+       p_number(atof(yytext));
-[0-9]*\.[0-9]+E[0-9]+      p_number(atof(yytext));
[0-9]+E[+-][0-9]+           p_number(atof(yytext));
-[0-9]+E[+-][0-9]+          p_number(atof(yytext));
[0-9]+E[0-9]+               p_number(atof(yytext));
-[0-9]+E[0-9]+              p_number(atof(yytext));
[0-9]*\.[0-9]+e[+-][0-9]+   p_number(atof(yytext));
-[0-9]*\.[0-9]+e[+-][0-9]+  p_number(atof(yytext));
[0-9]*\.[0-9]+e[0-9]+       p_number(atof(yytext));
-[0-9]*\.[0-9]+e[0-9]+      p_number(atof(yytext));
[0-9]+e[+-][0-9]+           p_number(atof(yytext));
-[0-9]+e[+-][0-9]+          p_number(atof(yytext));
[0-9]+e[0-9]+               p_number(atof(yytext));
-[0-9]+e[0-9]+              p_number(atof(yytext));
[0-9]*\.[0-9]*              p_number(atof(yytext));
-[0-9]*\.[0-9]*             p_number(atof(yytext));
[0-9]+                      p_int(atoi(yytext));
-[0-9]+                     p_int(atoi(yytext));
\+[0-9]+                    p_int(atoi(yytext));

\<\<                        p_token("<<");
\>\>                        p_token(">>");
\/\<\<                      p_quote("<<");
\/\>\>                      p_quote(">>");


[^ /\t\r\n(){}\[\]<>]+      p_token(yytext);

\/[^ \t\r\n(){}\[\]<>]*     p_quote(yytext+1);

\[                          p_token("[");
\]                          p_token("]");
\/\[                        p_quote("[");
\/\]                        p_quote("]");

%.*[\n]

[ \t\r\n]+

\(      string_buf_ptr = string_buf; BEGIN(str);

\{                          p_token("{");
\}                          p_token("}");

<str>\)        { /* saw closing quote - all done */
        BEGIN(INITIAL);
        *string_buf_ptr = '\0';
		p_string(string_buf);
        }

<str>\\[0-7]{3} {
        /* octal escape sequence */
        int result;

        (void) sscanf( yytext + 1, "%o", &result );

        if ( result > 0xff ) {
			yerror("Invalid octal escape sequence - constant is out of bounds");
			yyterminate();
		}

        *string_buf_ptr++ = result;
        }

<str>\\[0-9]+ {
		yerror("Invalid escape sequence");
        }

<str>\\n  *string_buf_ptr++ = '\n';
<str>\\t  *string_buf_ptr++ = '\t';
<str>\\r  *string_buf_ptr++ = '\r';
<str>\\b  *string_buf_ptr++ = '\b';
<str>\\f  *string_buf_ptr++ = '\f';

<str>\\[\n]

<str>\\(.|\n)  *string_buf_ptr++ = yytext[1];

<str>[^\\\n)]+        {
        char *yptr = yytext;

        while ( *yptr )
                *string_buf_ptr++ = *yptr++;
        }

