#ifndef _AREA_H
#define _AREA_H

#include "util.h"
#include "draw.h"

typedef struct {
	char *name;
	polygon_t *p;
} area_t;

typedef struct {
	unsigned int count;
	area_t *areas[];
} area_list_t;

area_list_t * area_list_new(int count);
int area_list_index(area_list_t *p, area_t *area);
area_list_t * area_list_copy(area_list_t *p);
area_list_t * area_list_add(area_list_t *p, area_t *area);
area_list_t * area_list_sub(area_list_t *p, area_t *area);
int area_list_eq(area_list_t *p1, area_list_t *p2);

ptrhash_t * area_list_hash_new();
vptrht_t * area_ptr_hash_new();

extern area_list_t **areamap;
extern int areamap_w, areamap_h;
extern double areamap_ox, areamap_oy, areamap_mx, areamap_my;

void init_area();
void kill_areas();
void set_area(char *name, polygon_t *p);
int warp_area(char *name);
void del_area(char *name);
void areamap_resize(int w, int h);
area_list_t *areamap_query(int x, int y);


void area_render_set(area_t *a);
void area_render_unset(area_t *a);
void render_areas();

/* retrieves list equivalent to parameter
 * from master hash
 * if it doesn't exist, list is copied and
 * put in hash */
area_list_t * area_list_store(area_list_t *l);

void clear_area_list_store();

area_list_t * area_list_store_add(area_list_t *p, area_t *area);
area_list_t * area_list_store_sub(area_list_t *p, area_t *area);

/* hash used for accelerating rendering:
 * area_list_t -> result area_list_t */
extern vptrht_t * area_render_hash;

void area_fillpoly(area_t *a, int set);

#endif
