#ifndef _FRTPLOT_GFX_H
#define _FRTPLOT_GFX_H

#include "SDL.h"
#include "poly.h"

extern SDL_Surface *main_surface;

/* size of a single character in the compiled in font */
extern int font_w, font_h;

/* line stipple counter */
extern unsigned int dashptr;

/* initialize gfx */
void gfx_init();

/* dot */
void gfx_dot(SDL_Surface *s, int x, int y, Uint32 color);
/* vertical line */
void gfx_vline(SDL_Surface *s, int x, int y, int y2, Uint32 color);
/* horizontal line */
void gfx_hline(SDL_Surface *s, int x1, int x2, int y, Uint32 color);
/* rectangle */
void gfx_rect(SDL_Surface *s, int x1, int y1, int x2, int y2, Uint32 color);
/* box */
inline void gfx_box(SDL_Surface *s, int x1, int y1, int x2, int y2, Uint32 color);
/* line */
void gfx_line(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color);
/* dashed line */
void gfx_line_dash(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint32 color, unsigned int dash);
/* draw a string */
void gfx_text(SDL_Surface *s, int x, int y, const char *str, const char *end, Uint8 alpha, Uint32 color);
/* fillpoly */
void gfx_fillpoly(SDL_Surface *surf, int n,int pt[][2],Uint8 alpha,Uint32 color);
void gfx_fillpoly_aa(SDL_Surface *surf, int n,int pt[][2],Uint8 alpha,Uint32 color);

void gfx_imgpoly(SDL_Surface *surf, int n,int pt[][2],Uint8 alpha, int ax, int ay, double xu, double xv, double yu, double yv, int imgalpha, SDL_Surface *img);

/* dot alpha */
void gfx_dot_alpha(SDL_Surface *s, int x, int y, Uint8 alpha, Uint32 color);
/* line with alpha */
void gfx_line_alpha(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 alpha, Uint32 color);
/* dashed line with alpha */
void gfx_line_alpha_dash(SDL_Surface * dst, Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Uint8 alpha, Uint32 color, unsigned int dash);
/* vertical line with alpha */
void gfx_vline_alpha(SDL_Surface *s, int x, int y, int y2, Uint8 alpha, Uint32 color);
/* horizontal line */
void gfx_hline_alpha(SDL_Surface *s, int x1, int x2, int y, Uint8 alpha, Uint32 color);
/* multalpha blit (pixel*given) */
void gfx_multalpha_blit(SDL_Surface *src, SDL_Surface *dst, int x, int y, Uint8 alpha);

typedef int (*gfx_polypt) [2];

void d_clear (Uint32 color);
void d_begin ();
void d_end ();
void d_clip (int x, int y, int w, int h);
void d_unclip ();
void d_fillpoly (int n, int pt[][2], Uint32 color);
Uint32 mapcolori(int r, int g, int b);

void d_ml_text(int x, int y, const char *str, Uint8 alpha, Uint32 color);
void d_line_thick(Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, float thick, Uint8 alpha, Uint32 color);
void d_line_thick_dash(Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, float thick, Uint8 alpha, Uint32 color, unsigned int dash);

Uint32 mapcolorf(float r, float g, float b);

/* use surface for drawing, use software routines
 * use this after your SDL_SetVideoMode call */
void gfx_surface_sw(SDL_Surface *s);

/* friendly clipping that watches for surface boundaries */
void gfx_clip(int x, int y, int w, int h);

/* increase this number every time colors need to be remapped */
extern Uint32 gfx_colorsupdate_serial;

/* structure for caching color mapping */
typedef struct {
	Uint32 rgb;    /* desired color */
	Uint32 clr;    /* mapped color */
	Uint32 update; /* last update serial */
} gfxclr;

/* return mapped version of color w/ caching */
static inline Uint32 gfx_useclr(gfxclr *clr)
{
	if(clr->update == gfx_colorsupdate_serial) return clr->clr;
	clr->update = gfx_colorsupdate_serial;
	clr->clr = mapcolori(clr->rgb >> 16, (clr->rgb >> 8) & 0xff, clr->rgb & 0xff);
	return clr->clr;
}

#define CLR(x) gfx_useclr(&(x))

static inline gfxclr mkclr(Uint32 clr)
{
	return (gfxclr) {
		.update = gfx_colorsupdate_serial - 1,
		.rgb = clr,
		.clr = 0
	};
}

static inline gfxclr mkclri(int r, int g, int b)
{
	return (gfxclr) {
		.update = gfx_colorsupdate_serial - 1,
		.rgb = (r << 16) | (g << 8) | b,
		.clr = 0
	};
}

static inline gfxclr mkclrf(float r, float g, float b)
{
	inline int cb(float x) {
		int z = x * 255.0;
		if(z < 0) z = 0;
		else if(z > 255) z = 255;
		return z;
	}
	int ir = cb(r);
	int ig = cb(g);
	int ib = cb(b);
	return (gfxclr) {
		.update = gfx_colorsupdate_serial - 1,
		.rgb = (ir << 16) | (ig << 8) | ib,
		.clr = 0
	};
}

extern unsigned char font_bits[];
extern int font_pitch;

/* horizontal line implementation 8bpp */
inline static void gfx_hline_8(SDL_Surface *s, int x1, int x2, int y, Uint8 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1;
	memset(p, color, (x2 - x1 + 1));
}

/* horizontal line implementation 16bpp */
inline static void gfx_hline_16(SDL_Surface *s, int x1, int x2, int y, Uint16 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1*2;
	Uint8 *p2 = p + (x2-x1+1)*2;
	while(p < p2) {
		*((Uint16*) p) = color;
		p += 2;
	}
}

/* horizontal line implementation 24bpp */
inline static void gfx_hline_24(SDL_Surface *s, int x1, int x2, int y, Uint32 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1*3;
	Uint8 *p2 = p + (x2-x1+1)*3;
	Uint8 c[3];
	if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
		c[0] = (color >> 16) & 0xff;
		c[1] = (color >> 8) & 0xff;
		c[2] = color & 0xff;
	} else {
		c[0] = color & 0xff;
		c[1] = (color >> 8) & 0xff;
		c[2] = (color >> 16) & 0xff;
	}

	while(p < p2) {
		*p++ = c[0];
		*p++ = c[1];
		*p++ = c[2];
	}
}

/* horizontal line implementation 32bpp */
inline static void gfx_hline_32(SDL_Surface *s, int x1, int x2, int y, Uint32 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1*4;
	Uint8 *p2 = p + (x2-x1+1)*4;
	while(p < p2) {
		*((Uint32*) p) = color;
		p += 4;
	}
}

/* horizontal line implementation 8bpp w/ alpha */
inline static void gfx_hline_alpha_8(SDL_Surface *s, int x1, int x2, int y, Uint8 alpha, Uint8 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1;
	Uint8 *p2 = p + (x2-x1+1);

	Uint32 ma = (alpha + 3) >> 5;
	Uint32 cg = (color & 0x1c) >> 2;
	Uint32 cng = color & 0xe3;
	Uint32 t1, t2;

	while(p < p2) {
		t1 = *p & 0xe3;
		t2 = *p & 0x1c;
		*p = (((((cg - (t2 >> 2)) * ma) >> 1) + t2) & 0x1c) | (((((cng - t1) * ma) >> 3) + t1) & 0xe3);
		p++;
	}
}

/* horizontal line implementation 15bpp w/ alpha */
inline static void gfx_hline_alpha_15(SDL_Surface *s, int x1, int x2, int y, Uint8 alpha, Uint16 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1*2;
	Uint8 *p2 = p + (x2-x1+1)*2;

	Uint32 ma = (alpha + 7) >> 3;
	Uint32 cg = (color & 0x03e0) >> 5;
	Uint32 cng = color & 0x7c1f;
	Uint32 t1, t2;

	while(p < p2) {
		t1 = *((Uint16*)p) & 0x7c1f;
		t2 = *((Uint16*)p) & 0x03e0;
		*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x03e0) | (((((cng - t1) * ma) >> 5) + t1) & 0x7c1f);
		p+=2;
	}
}

/* horizontal line implementation 16bpp w/ alpha */
inline static void gfx_hline_alpha_16(SDL_Surface *s, int x1, int x2, int y, Uint8 alpha, Uint16 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1*2;
	Uint8 *p2 = p + (x2-x1+1)*2;

	Uint32 ma = (alpha + 7) >> 3;
	Uint32 cg = (color & 0x07e0) >> 5;
	Uint32 cng = color & 0xf81f;
	Uint32 t1, t2;

	while(p < p2) {
		t1 = *((Uint16*)p) & 0xf81f;
		t2 = *((Uint16*)p) & 0x07e0;
		*((Uint16*)p) = ((((cg - (t2 >> 5)) * ma) + t2) & 0x07e0) | (((((cng - t1) * ma) >> 5) + t1) & 0xf81f);
		p+=2;
	}
}

/* horizontal line implementation 24bpp w/ alpha */
inline static void gfx_hline_alpha_24(SDL_Surface *s, int x1, int x2, int y, Uint8 alpha, Uint32 color)
{
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1*3;
	Uint8 *p2 = p + (x2-x1+1)*3;
	Uint32 c[3];
	Uint32 malpha = alpha;
	Uint32 ialpha = 255 - malpha;

	if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
		c[0] = ((color >> 16) & 0xff) * alpha;
		c[1] = ((color >> 8) & 0xff) * alpha;
		c[2] = (color & 0xff) * alpha;
	} else {
		c[0] = (color & 0xff) * alpha;
		c[1] = ((color >> 8) & 0xff) * alpha;
		c[2] = ((color >> 16) & 0xff) * alpha;
	}

	while(p < p2) {
		*p = (((Uint32) *p) * ialpha + c[0]) >> 8;
		p ++;
		*p = (((Uint32) *p) * ialpha + c[1]) >> 8;
		p ++;
		*p = (((Uint32) *p) * ialpha + c[2]) >> 8;
		p ++;
	}
}

/* horizontal line implementation 32bpp w/ alpha */
inline static void gfx_hline_alpha_32(SDL_Surface *s, int x1, int x2, int y, Uint8 alpha, Uint32 color)
{
#ifdef BGR
	color = SDL_Swap32(color);
#endif
	Uint8 *p = ((Uint8 *)s->pixels) + s->pitch * y + x1*4;
	Uint8 *p2 = p + (x2-x1+1)*4;
	Uint32 cg = color & 0x00ff00;
	Uint32 cng = color & 0xff00ff;
	Uint32 t1, t2;
	Uint32 ip;
	while(p < p2) {
#ifdef BGR
		ip = SDL_Swap32(*((Uint32*) p));
#else
		ip = *((Uint32*) p);
#endif
		t2 = ip & 0xff00ff;
		t1 = cng - t2;
		t1 = (((t1*alpha) >> 8) + t2) & 0xff00ff;
		ip &= 0xff00;
		t2 = ((((cg - ip) * alpha) >> 8) + ip) & 0xff00;
#ifdef BGR
		*((Uint32*) p) = SDL_Swap32(t1 | t2);
#else
		*((Uint32*) p) = t1 | t2;
#endif
		p += 4;
	}
}



#endif
