#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "SDL.h"
#include "poly.h"

#define imul32(X,Y) ((int)(X) * (int)(Y))
#define iscale(X,N,D)   (int)(imul32(X,N) / (int)(D))
#define bitsof(s)   (sizeof(s) * 8)

#define iswap(x,y) {                                            \
        int _swap_tmpval_ = (x);                                \
        (x) = (y);                                              \
        (y) = _swap_tmpval_;                                    \
}

#define isort(x,y) {                                            \
        if((int)(x) > (int)(y)) iswap(x,y)                      \
}

#define irscale(X,N,D)  ((                                              \
        iscale(((int)(X) << 1),N,D) +                                   \
        (((int)(X) ^ (int)(N) ^ (int)(D)) >> (bitsof(int) - 1)) +       \
        1                                                               \
) >> 1)

#define clip_line_ymin(x1,y1,x2,y2) {                           \
	if(y1 < lowy) {                                             \
		x1 += irscale((x2 - x1),(lowy - y1),(y2 - y1));         \
		y1  = lowy;                                             \
	}                                                           \
}

#define clip_line_ymin_a(x1,y1,x2,y2) {                         \
	if(y1 < mlowy) {                                            \
		x1 += irscale((x2 - x1),(mlowy - y1),(y2 - y1));        \
		y1  = mlowy;                                            \
	}                                                           \
}

#define clip_ordxrange_(x1,x2,when_out) {                       \
	if(x1 > highx) { when_out; }                                \
	if(x2 < lowx) { when_out; }                                 \
	if(x1 < lowx) { x1 = lowx; }                                \
	if(x2 > highx) { x2 = highx; }                              \
}

#define TYPE int
#define LTE(a,b) ((a) <= (b))
#define CALL intsort
#include "shellsort.h"
#undef TYPE
#undef LTE
#undef CALL

/* edge of a polygon */
typedef struct {
	int x,xlast;                /* current and final X coordinates */
	int y,ylast;                /* current and final Y coordinates */
	int dx,dy;                  /* line diagonals */
	int xmajor;                 /* flag for X major lines */
	int xstep;                  /* direction of X scan */
	int error;                  /* Bresenham error term */

	int active;                 /* index in active edge table */
	int flipped;                /* whether edge is y-flipped */
} polyedge_t;

/* edge event */
typedef struct {
	int ypos;
	unsigned short enter : 1;
	unsigned short horiz : 1;
	unsigned int edge : 30;
} edgevent_t;

#define TYPE edgevent_t
#define LTE(a,b) ((a).ypos < (b).ypos)
#define CALL edgeventsort
#include "shellsort.h"
#undef TYPE
#undef LTE
#undef CALL

/* calculate edge bresenham parameters based on x,y -> xlast,ylast */
static void setup_edge(polyedge_t *ep)
{
	ep->dy = ep->ylast - ep->y;
	ep->dx = ep->xlast - ep->x;
	if(ep->dx < 0) {
		ep->xstep = (-1);
		ep->dx    = (-ep->dx);
	}
	else {
		ep->xstep  = 1;
	}
	if(ep->dx > ep->dy) {
		ep->xmajor = 1;
		ep->error  = ep->dx >> 1;
	}
	else {
		ep->xmajor = 0;
		ep->error  = (ep->dy - ((1 - ep->xstep) >> 1)) >> 1;
	}
}

/* x step for x-major edge */
static inline void xstep_edge(polyedge_t *ep)
{
	for( ; ; ) {
		ep->x += ep->xstep;
		if((ep->error -= ep->dy) < 0) {
			ep->error += ep->dx;
			break;
		}
	}
}

/* x step for y-major edge */
static inline void ystep_edge(polyedge_t *ep)
{
	if((ep->error -= ep->dx) < 0) {
		ep->x     += ep->xstep;
		ep->error += ep->dy;
	}
}

/* make x step */
static inline void step_edge(polyedge_t *ep)
{
	if(ep->xmajor) xstep_edge(ep); else ystep_edge(ep);
}

/* edge table */
static polyedge_t *edges = NULL;
static int edges_size = 0;

/* edge event table */
static edgevent_t *edgevents = NULL;
static int edgevents_size = 0;

/* active edge table (size = edges_size) */
static polyedge_t **actedges = NULL;

/* point table (size = edgevents_size = edges_size*2) */
static int *points = NULL;

/* segment table (2 segments per point) */
static poly_hline_segment_t *segments = NULL;


void scanpoly(int lowx, int highx, int lowy, int highy, int n, int pt[][2], poly_hline_t hline)
{
	if((n > 1) &&
			(pt[0][0] == pt[n-1][0]) &&
			(pt[0][1] == pt[n-1][1])) {
		n--;
	}
	if(n < 1) return;

	/* check that data structures are big enough */
	if(edges_size < n + 2) {
		edges_size = n + 2;
		edges = realloc(edges, edges_size * sizeof *edges);
		if(!edges) abort();
		actedges = realloc(actedges, edges_size * sizeof *actedges);
		if(!actedges) abort();
	}

	if(edgevents_size < edges_size * 2) {
		edgevents_size = edges_size * 2;
		edgevents = realloc(edgevents, edgevents_size * sizeof *edgevents);
		if(!edgevents) abort();
		points = realloc(points, 2 * POLY_A_YMULT * edgevents_size * sizeof *points);
		if(!points) abort();
		segments = realloc(segments, 2 * POLY_A_YMULT * edgevents_size * sizeof *segments);
		if(!segments) abort();
	}

	int i, xmin, xmax, ymin, ymax, prevx, prevy, nedges;
	polyedge_t *ep;

	prevx = xmax = xmin = pt[0][0];
	prevy = ymax = ymin = pt[0][1];

	nedges = 0;
	ep = edges;

	/* build edge table */
	i = n;
	while(--i >= 0) {
		if(pt[i][1] >= prevy) {
			ep->x = prevx;
			ep->y = prevy;
			ep->xlast = prevx = pt[i][0];
			ep->ylast = prevy = pt[i][1];
			ep->flipped = 0;
		} else {
			ep->xlast = prevx;
			ep->ylast = prevy;
			ep->x = prevx = pt[i][0];
			ep->y = prevy = pt[i][1];
			ep->flipped = 1;
		}

		if(ep->y > highy || ep->ylast < lowy) continue;
		clip_line_ymin(ep->x, ep->y, ep->xlast, ep->ylast);
		if(ymin > ep->y) ymin = ep->y;
		if(ymax < ep->ylast) ymax = ep->ylast;
		if(xmin > ep->x) xmin = ep->x;
		if(xmax < ep->xlast) xmax = ep->xlast;
		setup_edge(ep);
		nedges ++;
		ep ++;
	}

	if(!nedges || xmin > highx || xmax < lowx) return;

	/* generate edge event table */
	int nedgevents = 0;
	edgevent_t *ev;

	for(i=0, ep=edges, ev=edgevents; i<nedges; i++, ep++) {
		if(ep->y != ep->ylast) {
			ev->ypos = ep->y;
			ev->enter = 1;
			ev->edge = i;
			ev ++; nedgevents ++;

			ev->ypos = ep->ylast;
			ev->enter = 0;
			ev->edge = i;
			ev ++; nedgevents ++;
		}

		//printf("edge %d %d,%d->%d,%d flipped %d\n", i, ep->x, ep->y, ep->xlast, ep->ylast, ep->flipped);
	}

	/* sort edge event table */
	edgeventsort(edgevents, nedgevents);

	/* number of active edges */
	int nactive = 0;

	/* helper function: activate edge */
	inline void edgeon(polyedge_t *ep)
	{
		actedges[nactive] = ep;
		ep->active = nactive;
		nactive ++;
	}

	/* helper function: deactivate edge */
	inline void edgeoff(polyedge_t *ep)
	{
		actedges[ep->active] = actedges[nactive - 1];
		actedges[nactive - 1]->active = ep->active;
		nactive --;
	}

	if(ymin < lowy) ymin = lowy;
	if(ymax > highy) ymax = highy;

	edgevent_t *maxev = edgevents + nedgevents;
	ev = edgevents;

	/* process events below ymin */
	while(ev < maxev && ev->ypos < ymin) {
		if(ev->enter)
			edgeon(edges + ev->edge);
		else
			edgeoff(edges + ev->edge);

		ev ++;
	}

	/* scan polygon */
	int ypos;
	polyedge_t **aep;
	int *cpt, *cpt1;
	for(ypos = ymin; ypos <= ymax; ypos++) {
		cpt = points;

		/* handle events */
		while(ev < maxev && ev->ypos == ypos) {
			ep = edges + ev->edge;
			if(ev->enter) edgeon(ep); else edgeoff(ep);
			ev ++;
		}

		/* process active edges */
		for(i=0, aep=actedges; i<nactive; i++, aep++) {
			*cpt++ = (*aep)->x;
			step_edge(*aep);
		}

		/* sort generated points */
		intsort(points, cpt - points);

		assert(!((cpt - points) & 1)); /* number of generated points should be even */

		/* scan segments */
		for(cpt1 = points; cpt1 < cpt; cpt1 += 2) {
			int x1 = cpt1[0];
			int x2 = cpt1[1];

			/* merge continuous segments */
			while(cpt1 < cpt-2 && x2 == cpt1[2]) {
				x2 = cpt1[3];
				cpt1 += 2;
			}

			x2 --;
			clip_ordxrange_(x1, x2, continue);
			hline(x1, x2, ypos);
		}
	}
}

void scanpoly_a(int lowx, int highx, int lowy, int highy, int n, int pt[][2], poly_hline_a_t cb, void *data)
{
	/* bounds in premult coordinates */
	int mlowx = lowx * POLY_A_XMULT;
	int mlowy = lowy * POLY_A_YMULT;
	int mhighx = (highx+1) * POLY_A_XMULT - 1;
	int mhighy = (highy+1) * POLY_A_YMULT - 1;

	if((n > 1) &&
			(pt[0][0] == pt[n-1][0]) &&
			(pt[0][1] == pt[n-1][1])) {
		n--;
	}
	if(n < 1) return;

	/* check that data structures are big enough */
	if(edges_size < n + 2) {
		edges_size = n + 2;
		edges = realloc(edges, edges_size * sizeof *edges);
		if(!edges) abort();
		actedges = realloc(actedges, edges_size * sizeof *actedges);
		if(!actedges) abort();
	}

	if(edgevents_size < edges_size * 2) {
		edgevents_size = edges_size * 2;
		edgevents = realloc(edgevents, edgevents_size * sizeof *edgevents);
		if(!edgevents) abort();
		points = realloc(points, POLY_A_YMULT * edgevents_size * sizeof *points);
		if(!points) abort();
		segments = realloc(segments, 2 * POLY_A_YMULT * edgevents_size * sizeof *segments);
		if(!segments) abort();
	}

	int i, xmin, xmax, ymin, ymax, prevx, prevy, nedges;
	polyedge_t *ep;

	prevx = xmax = xmin = pt[0][0];
	prevy = ymax = ymin = pt[0][1];

	nedges = 0;
	ep = edges;

	/* build edge table */
	i = n;
	while(--i >= 0) {
		if(pt[i][1] >= prevy) {
			ep->x = prevx;
			ep->y = prevy;
			ep->xlast = prevx = pt[i][0];
			ep->ylast = prevy = pt[i][1];
			ep->flipped = 0;
		} else {
			ep->xlast = prevx;
			ep->ylast = prevy;
			ep->x = prevx = pt[i][0];
			ep->y = prevy = pt[i][1];
			ep->flipped = 1;
		}

		if(ep->y > mhighy || ep->ylast < mlowy) continue;
		clip_line_ymin_a(ep->x, ep->y, ep->xlast, ep->ylast);
		if(ymin > ep->y) ymin = ep->y;
		if(ymax < ep->ylast) ymax = ep->ylast;
		if(xmin > ep->x) xmin = ep->x;
		if(xmax < ep->xlast) xmax = ep->xlast;
		setup_edge(ep);
		nedges ++;
		ep ++;
	}

	if(!nedges || xmin > mhighx || xmax < mlowx) return;

	/* generate edge event table */
	int nedgevents = 0;
	edgevent_t *ev;

	for(i=0, ep=edges, ev=edgevents; i<nedges; i++, ep++) {
		if(ep->y != ep->ylast) {
			ev->ypos = ep->y;
			ev->enter = 1;
			ev->edge = i;
			ev ++; nedgevents ++;

			ev->ypos = ep->ylast;
			ev->enter = 0;
			ev->edge = i;
			ev ++; nedgevents ++;
		}

		//printf("edge %d %d,%d->%d,%d flipped %d\n", i, ep->x, ep->y, ep->xlast, ep->ylast, ep->flipped);
	}

	/* sort edge event table */
	edgeventsort(edgevents, nedgevents);

	/* number of active edges */
	int nactive = 0;

#define EDGEON(ep) \
	actedges[nactive] = ep; \
	(ep)->active = nactive; \
	nactive ++;

#define EDGEOFF(ep) \
	actedges[(ep)->active] = actedges[nactive - 1]; \
	actedges[nactive - 1]->active = (ep)->active; \
	nactive --;

#define DISPATCHSEG \
	/* printf("dispatching: %d .. %d  alpha %d\n", seg->x1, seg->x2, seg_alpha); */ \
	if(seg->x1 < lowx) seg->x1 = lowx; \
	if(seg->x2 > highx) seg->x2 = highx; /* XXX: off by one? */ \
	if(seg->x2 > seg->x1) seg ++;

	if(ymin < mlowy) ymin = mlowy;
	if(ymax > mhighy) ymax = mhighy;

	edgevent_t *maxev = edgevents + nedgevents;
	ev = edgevents;

	/* process events below ymin */
	while(ev < maxev && ev->ypos < ymin) {
		if(ev->enter) {
			EDGEON(edges + ev->edge);
		} else {
			EDGEOFF(edges + ev->edge);
		}

		ev ++;
	}

	/* calculate scanline min, max */
	int pymin = ymin / POLY_A_YMULT;
	int pymax = ymax / POLY_A_YMULT;

	/* scan polygon, scanline by scanline */
	int pypos;
	polyedge_t **aep;
	int *cpt;

	int ypos = ymin;

	for(pypos = pymin; pypos <= pymax; pypos++) {
		cpt = points;

		int ynext = (pypos + 1) * POLY_A_YMULT;

		/* process logical lines, generate points.
		 * point x coordinates are shifted left by POLY_A_YSHIFT
		 * and the created space will store the scanline offset.
		 * the resulting array is still sortable by X */
		for(; ypos < ynext; ypos ++) {
			/* handle events */
			while(ev < maxev && ev->ypos == ypos) {
				ep = edges + ev->edge;
				if(ev->enter) {
					EDGEON(ep);
				} else {
					EDGEOFF(ep);
				}
				ev ++;
			}

			//printf("ypos=%d\n", ypos);

			/* process active edges */
			for(i=0, aep=actedges; i<nactive; i++, aep++) {
				int off = ynext - ypos - 1; /* 0 .. (POLY_A_YMULT-1) */

				*cpt++ = (((*aep)->x) << POLY_A_YSHIFT) + off;
				//printf("      x=%d point=%d\n", (*aep)->x, cpt[-1]);
				step_edge(*aep);
			}
		}

		/* sort generated points */
		intsort(points, cpt - points);

		assert(!((cpt - points) & 1)); /* number of generated points should be even */

		poly_hline_segment_t *seg = segments;
		char sbits[POLY_A_YMULT]; /* bools for each scanline */
		int scount = 0;           /* count of trues in sbits */

		int i;
		for(i=0; i<POLY_A_YMULT; i++) sbits[i] = 0;

		/* fill scanline */

		int prev_x = xmin;

		/* unfinished segment */
		int seg_px1 = prev_x / POLY_A_XSHIFT;  /* first (full) pixel */
		int seg_alpha = 0;

		int pix_fill = 0; /* current pixel fill, xmult * ymult */

		/* algo:
		 * each processed point is a fill (alpha) change at some X.
		 *
		 * we maintain a full draw segment (that begins and ends at full
		 * pixels) and has a given fill.
		 *
		 * the "current pixel" is right after the draw segment, at prev_x
		 * (/ POLY_A_XSHIFT).
		 *
		 * we're always working on the current pixel separately from the
		 * unfinished segment (this is in pix_fill)
		 *
		 * each processed point draws a logical segment from prev_x to x.
		 *
		 * the beginning of the logical segment always falls into the
		 * current pixel. first, update current pixel.
		 * if the end of the logical segment is >= current pixel, then
		 * check if the current pixel can be appended to the segment.
		 * if yes, just append. if not, dispatch the segment, and the current
		 * pixel becomes an 1 pixel wide draw segment.
		 *
		 * by now the draw segment includes the end of the (finished)
		 * current pixel.
		 *
		 * if x is larger than the end of the pixel after the current pixel
		 * (prev_x_px + 1), then generate/compare a new segment.
		 *
		 * set up portion of segment in the last pixel (x_px .. x) as the
		 * current pixel
		 *
		 */


		int *ipt;
		for(ipt = points; ipt < cpt; ipt ++) {
			int x = (*ipt) >> POLY_A_YSHIFT;
			int off = *ipt & (POLY_A_YMULT - 1);

			int x_px = x >> POLY_A_XSHIFT;
			int prev_x_px = prev_x >> POLY_A_XSHIFT;

			/* x at the end of the current pixel, or at the end of the
			 * segment - whichever shorter */
			int pix_end_x = (prev_x_px + 1) << POLY_A_XSHIFT;
			if(x < pix_end_x) pix_end_x = x;

			/* update current pixel */
			pix_fill += (pix_end_x - prev_x) * scount;

			//printf("x %d off %d scount(b) %d x_px %d pix_end_x %d fill %d\n", x, off, scount, x_px, pix_end_x, pix_fill);

			if(x_px > prev_x_px) {
				/* finished with the current pixel */

				/* pix_fill can go from 0 (alpha=0) to
				 * POLY_A_YMULT * POLY_A_XMULT (alpha=255) */
				int alpha = pix_fill * 255 / (POLY_A_XMULT * POLY_A_YMULT);

				//printf("     pixel over, alpha %d\n", alpha);

				if(alpha != seg_alpha) {
					/* dispatch segment until current pixel */
					if(seg_alpha > 0 && prev_x_px > seg_px1) {
						seg->x1 = seg_px1;
						seg->x2 = prev_x_px;
						seg->alpha = seg_alpha;
						DISPATCHSEG;
					}

					seg_px1 = prev_x_px;
					seg_alpha = alpha;

					//printf("    (1) segment at: px1 %d alpha %d\n", seg_px1, seg_alpha);
				}

				if(x_px > prev_x_px + 1) {
					/* generate new segment */
					alpha = scount * 255 / POLY_A_YMULT;

					//printf("    generating new segment, alpha %d x_px %d seg_px1 %d\n", alpha, x_px, seg_px1);

					if(alpha != seg_alpha) {
						if(seg_alpha > 0) {
							seg->x1 = seg_px1;
							seg->x2 = prev_x_px + 1;
							seg->alpha = seg_alpha;
							DISPATCHSEG;
						}

						seg_px1 = prev_x_px + 1;
						seg_alpha = alpha;

						//printf("    (2) segment at: px1 %d alpha %d\n", seg_px1, seg_alpha);
					}
				}

				/* next current pixel is the part from x_px .. x */
				pix_fill = scount * (x - (x_px << POLY_A_XSHIFT));
			}

			if((sbits[off] ^= 1)) {
				scount ++;
			} else {
				scount --;
			}

			prev_x = x;
		}

		assert(scount == 0);

		//printf("final fill %d\n", pix_fill);

		int prev_x_px = prev_x >> POLY_A_XSHIFT;

		int alpha = pix_fill * 255 / (POLY_A_XMULT * POLY_A_YMULT);

		if(alpha != seg_alpha) {
			/* dispatch segment until current pixel */
			if(seg_alpha > 0 && prev_x_px > seg_px1) {
				seg->x1 = seg_px1;
				seg->x2 = prev_x_px;
				seg->alpha = seg_alpha;
				DISPATCHSEG;
			}

			seg_px1 = prev_x_px;
			seg_alpha = alpha;

			//printf("    (3) segment at: px1 %d alpha %d\n", seg_px1, seg_alpha);
		}

		int final_x = prev_x_px + 1;
		if(seg_alpha > 0 && final_x > seg_px1) {
			//printf("final seg_alpha %d seg_px1 %d\n", seg_alpha, seg_px1);
			seg->x1 = seg_px1;
			seg->x2 = final_x;
			seg->alpha = seg_alpha;
			DISPATCHSEG;
		}

		cb(pypos, segments, seg, data);
	}
}

/*
int main(int argc, char *argv[])
{
//	int pt[][2] = {
//		{0, 0}, {10, 0}, {10, 10}, {0, 10}, {0, 0},
//		{3, 3}, {7, 3}, {7, 7}, {3, 7}, {3, 3},
//	};
	
//	int pt[][2] = {
//		{5, 0}, {15, 0}, {15, 10}, {10, 10}, {5, 15},
//		{0, 10}, {5, 5}, {10, 10}, {5, 10}
//	};

//	int pt[][2] = {
//		{0,0}, {10,0}, {10,5}, {20,5}, {10,5},
//		{10, 10}, {0,10}
//	};
	
//	int pt[][2] = {
//		{0,0}, {10,0}, {10,5}, {10,10}, {0,10}
//	};

//	int pt[][2] = {
//		{10,0}, {20, 10}, {10, 20}, {0, 10}
//	};

	int pt[][2] = {
		{47,58}, {169, 30}, {117, 38}, {162, 34}, {97, 30}
	};

	//4.75575 5.82132
		//16.9163 3.04416
		//11.7108 3.8695
		//16.2125 3.47062
		//9.69966 3.03726

	int i;
	for(i=0; i<5; i++) {
		pt[i][0] *= POLY_A_XMULT;
		pt[i][1] *= POLY_A_YMULT;
	}

	void hline(int y, poly_hline_segment_t *i, poly_hline_segment_t *end, void *data)
	{
		printf("hline y=%d\n", y);
		for(; i<end; i++) {
			printf("    %4d .. %4d       %d\n", i->x1, i->x2, i->alpha);
		}
	}

	scanpoly_a(0, 180, 0, 70, 5, pt, hline, NULL);

	return 0;
}
*/
