#ifndef _VPTRHASH_H
#define _VPTRHASH_H

/* open addressing hashtable: void* -> void* */
/* no memory management for keys and values */
/* no malloc checks (out of memory == segfault) */
/* limitation: cannot delete items, value cannot be 0 */

/* c89 compatibility */
#ifdef __STRICT_ANSI__
#define inline
#endif

typedef void *vptrht_key_t;
typedef void *vptrht_value_t;
typedef unsigned int vptrht_hash_t;

typedef struct {
	vptrht_key_t key;
	vptrht_value_t value;
} vptrht_entry_t;

typedef struct {
	unsigned int mask;
	unsigned int fill;
	vptrht_entry_t *table;
} vptrht_t;

vptrht_t *vptrht_alloc(void);
void vptrht_init(vptrht_t *ht);
void vptrht_free(vptrht_t *ht);
void vptrht_uninit(vptrht_t *ht);
void vptrht_clear(vptrht_t *ht);
/* new size is 2^n >= hint, abort() if maxsize is reached */
void vptrht_resize(vptrht_t *ht, unsigned int hint);

/* key is in ht */
int vptrht_has(vptrht_t *ht, vptrht_key_t key);
/* ht[key] or 0 if key is not in ht */
vptrht_value_t vptrht_get(vptrht_t *ht, vptrht_key_t key);
/* ht[key] = value */
void vptrht_set(vptrht_t *ht, vptrht_key_t key, vptrht_value_t value);

/* helper functions */
static inline unsigned int vptrht_length(const vptrht_t *ht) {return ht->fill;}
static inline unsigned int vptrht_capacity(const vptrht_t *ht) {return ht->mask + 1;}

/* ht cannot hold 0 keys */
static inline int vptrht_isempty(vptrht_entry_t *e) {return e->key == 0;}

/* macro iteration: vptrht_entry_t *e */
#define vptrht_foreach(ht, e) \
	for (e = ht->table; e != ht->table + ht->mask + 1; e++) \
		if (!vptrht_isempty(e))

#endif
