#!/bin/sh
if [ -z "$1" ]; then
	echo "Usage: $0 drawing"
	exit 1
fi
exec plumb -V ctrl awk -v file="$1" '
function redraw()
{
	print "frame"
	print "ident"
	print "shift " (-shiftx) " " (-shifty)
	print "scale " zoom
	if(macro) {
		print data
	} else {
		print "invoke pic"
	}
	print "flush"
	fflush()
}

BEGIN {
	drag=0
	origx=0
	origy=0
	oshiftx=0
	oshifty=0
	shiftx=0
	shifty=0
	zoom=1
	macro=0
	lines=0

	while((getline < file) == 1) {
		if(/flush/) {
			print "events on"
			if(!macro) {
				print "macro pic"
				for(n=0; n<lines; n++)
					print data[n];
				print "endmacro"
			}
			redraw()
			break;
		} else if(/^bg /) {
			print $0
		} else {
			data[lines] = $0 "\n";
			lines++;
			if(/macro/)
				macro = 1;
		}
	}
}

/PRESS/ {
	if($2 == 3) {
		drag=1;
		origx=$3;
		origy=$4;
		oshiftx=shiftx;
		oshifty=shifty;
	} else if($2 == 4) {
		zoom=zoom*1.1
		shiftx=shiftx*1.1
		shifty=shifty*1.1
	} else if($2 == 5) {
		zoom=zoom/1.1
		shiftx=shiftx/1.1
		shifty=shifty/1.1
	} else if($2 == 1) {
		shiftx=0; shifty=0; zoom=1;
	}
	redraw()
}

/RELEASE/ {
	if($2 == 3) {
		drag=0;
	}
}

/MOTION/ {
	if(drag) {
		shiftx = oshiftx - ($3 - origx);
		shifty = oshifty - ($4 - origy);
		redraw();
	}
}

' . '
proc ctrl = $ctrl
proc anim = animator -q
pipe ctrl anim
pipe anim ctrl
'
